@@ -2,21 +2,19 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-import json
-import pickle
-# import schema
 from collections import OrderedDict
-from copy import deepcopy
 from ast import literal_eval
 from random import sample
 from random import choice
+from distutils.version import LooseVersion
 
-# import compas
-from compas.datastructures.mesh.core import VertexAttributeView
-from compas.datastructures.mesh.core import EdgeAttributeView
-from compas.datastructures.mesh.core import FaceAttributeView
+import compas
+
+from ...datastructure import Datastructure
+from ...attributes import VertexAttributeView
+from ...attributes import EdgeAttributeView
+from ...attributes import FaceAttributeView
 
-from compas.datastructures import Datastructure
 from compas.utilities import pairwise
 from compas.utilities import window
 
@@ -29,69 +27,128 @@ class HalfEdge(Datastructure):
 
     Attributes
     ----------
-    attributes
-    default_vertex_attributes
-    default_edge_attributes
-    default_face_attributes
-    name
-    adjacency
-    data
+    attributes : dict
+        Named attributes related to the data structure as a whole.
+    default_vertex_attributes : dict
+        Named attributes and default values of the vertices of the data structure.
+    default_edge_attributes : dict
+        Named attributes and default values of the edges of the data structure.
+    default_face_attributes : dict
+        Named attributes and default values of the faces of the data structure.
+    name : str
+        Name of the data structure.
+        Defaults to the value of `self.__class__.__name__`.
+    data : dict
+        The data representation of the data structure.
+    adjacency : dict
+        Alias for `self.halfedge`.
+
+        .. deprecated:: 0.17.0
 
-    Examples
-    --------
-    >>>
     """
 
-    # DATASCHEMA = schema.Schema({
-    #     "compas": str,
-    #     "datatype": str,
-    #     "data": {
-    #         "attributes": dict,
-    #         "dva": dict,
-    #         "dea": dict,
-    #         "dfa": dict,
-    #         "vertex": dict,
-    #         "face": dict,
-    #         "facedata": dict,
-    #         "edgedata": dict,
-    #         "max_int_key": schema.And(int, lambda x: x >= -1),
-    #         "max_int_fkey": schema.And(int, lambda x: x >= -1)
-    #     }
-    # })
+    @property
+    def DATASCHEMA(self):
+        import schema
+        if LooseVersion(compas.__version__) < LooseVersion('0.16.5'):
+            return schema.Schema({
+                "attributes": dict,
+                "dva": dict,
+                "dea": dict,
+                "dfa": dict,
+                "vertex": schema.And(
+                    dict,
+                    lambda x: all(isinstance(key, int) for key in x)
+                ),
+                "face": schema.And(
+                    dict,
+                    lambda x: all(isinstance(key, int) for key in x),
+                    lambda x: all(all(isinstance(item, int) for item in value) for value in x.values())
+                ),
+                "facedata": schema.And(
+                    dict,
+                    lambda x: all(isinstance(key, int) for key in x),
+                    lambda x: all(isinstance(value, dict) for value in x.values())
+                ),
+                "edgedata": dict,
+                "max_int_key": schema.And(int, lambda x: x >= -1),
+                "max_int_fkey": schema.And(int, lambda x: x >= -1)
+            })
+        return schema.Schema({
+            "compas": str,
+            "datatype": str,
+            "data": {
+                "attributes": dict,
+                "dva": dict,
+                "dea": dict,
+                "dfa": dict,
+                "vertex": schema.And(
+                    dict,
+                    lambda x: all(isinstance(key, int) for key in x)
+                ),
+                "face": schema.And(
+                    dict,
+                    lambda x: all(isinstance(key, int) for key in x),
+                    lambda x: all(all(isinstance(item, int) for item in value) for value in x.values())
+                ),
+                "facedata": schema.And(
+                    dict,
+                    lambda x: all(isinstance(key, int) for key in x),
+                    lambda x: all(isinstance(value, dict) for value in x.values())
+                ),
+                "edgedata": schema.And(
+                    dict,
+                    lambda x: all(isinstance(key, str) for key in x),
+                    lambda x: all(isinstance(value, dict) for value in x.values())
+                ),
+                "max_vertex": schema.And(int, lambda x: x >= -1),
+                "max_face": schema.And(int, lambda x: x >= -1)
+            }
+        })
 
-    # JSONSCHEMA = {
-    #     "$schema": "http://json-schema.org/schema",
-    #     "$id": "https://github.com/compas-dev/compas/schemas/mesh.json",
-    #     "$compas": compas.__version__,
-
-    #     "type": "object",
-    #     "poperties": {
-    #         "compas": {"type": "string"},
-    #         "datatype": {"type": "string"},
-    #         "data": {
-    #             "type": "object",
-    #             "properties": {
-    #                 "attributes":   {"type": "object"},
-    #                 "dva":          {"type": "object"},
-    #                 "dea":          {"type": "object"},
-    #                 "dfa":          {"type": "object"},
-    #                 "vertex":       {"type": "object"},
-    #                 "face":         {"type": "object"},
-    #                 "facedata":     {"type": "object"},
-    #                 "edgedata":     {"type": "object"},
-    #                 "max_int_key":  {"type": "number"},
-    #                 "max_int_fkey": {"type": "number"}
-    #             },
-    #             "required": ["attributes", "dva", "dea", "dfa", "vertex", "face", "facedata", "edgedata", "max_int_key", "max_int_fkey"]
-    #         }
-    #     },
-    #     "required": ["compas", "datatype", "data"]
-    # }
+    @property
+    def JSONSCHEMA(self):
+        version = LooseVersion(compas.__version__)
+        schema = {
+            "$schema": "http://json-schema.org/draft-07/schema#",
+            "$id": "https://github.com/compas-dev/compas/schemas/halfedge.json",
+            "$compas": version.vstring.split('-')[0],
+        }
+        data = {
+            "type": "object",
+            "properties": {
+                "attributes":   {"type": "object"},
+                "dva":          {"type": "object"},
+                "dea":          {"type": "object"},
+                "dfa":          {"type": "object"},
+                "vertex":       {"type": "object"},
+                "face":         {"type": "object"},
+                "facedata":     {"type": "object"},
+                "edgedata":     {"type": "object"},
+                "max_vertex":   {"type": "number"},
+                "max_face":     {"type": "number"}
+            },
+            "required": ["attributes", "dva", "dea", "dfa", "vertex", "face", "facedata", "edgedata", "max_vertex", "max_face"]
+        }
+        if version < LooseVersion('0.16.5'):
+            schema.update(data)
+        else:
+            meta = {
+                "type": "object",
+                "properties": {
+                    "compas": {"type": "string"},
+                    "datatype": {"type": "string"},
+                    "data": data
+                },
+                "required": ["compas", "datatype", "data"]
+            }
+            schema.update(meta)
+        return schema
 
     def __init__(self):
         super(HalfEdge, self).__init__()
-        self._max_int_key = -1
-        self._max_int_fkey = -1
+        self._max_vertex = -1
+        self._max_face = -1
         self.vertex = {}
         self.halfedge = {}
         self.face = {}
@@ -103,28 +160,7 @@ class HalfEdge(Datastructure):
         self.default_face_attributes = {}
 
     # --------------------------------------------------------------------------
-    # customisation
-    # --------------------------------------------------------------------------
-
-    def __str__(self):
-        """Generate a readable representation of the data of the mesh."""
-        return json.dumps(self.data, sort_keys=True, indent=4)
-
-    def summary(self):
-        """Print a summary of the mesh."""
-        tpl = "\n".join(
-            ["Mesh summary",
-             "============",
-             "- vertices: {}",
-             "- edges: {}",
-             "- faces: {}"])
-        s = tpl.format(self.number_of_vertices(),
-                       self.number_of_edges(),
-                       self.number_of_faces())
-        print(s)
-
-    # --------------------------------------------------------------------------
-    # special properties
+    # descriptors
     # --------------------------------------------------------------------------
 
     @property
@@ -147,235 +183,109 @@ class HalfEdge(Datastructure):
     @property
     def data(self):
         """dict : A data dict representing the mesh data structure for serialisation.
-
-        The dict has the following structure:
-
-        * 'attributes'   => dict
-        * 'dva'          => dict
-        * 'dea'          => dict
-        * 'dfa'          => dict
-        * 'vertex'       => dict
-        * 'face'         => dict
-        * 'facedata'     => dict
-        * 'edgedata'     => dict
-        * 'max_int_key'  => int
-        * 'max_int_fkey' => int
-
-        Notes
-        -----
-        All dictionary keys are converted to their representation value (``repr(key)``)
-        to ensure compatibility of all allowed key types with the JSON serialisation
-        format, which only allows for dict keys that are strings.
-
         """
-        # vertex = {}
-        # face = {}
-        # facedata = {}
-        edgedata = {}
-
-        # for key in self.vertex:
-        #     vertex[repr(key)] = self.vertex[key]
-
-        # for key in self.face:
-        #     face[repr(key)] = [repr(k) for k in self.face[key]]
-
-        # for key in self.facedata:
-        #     facedata[repr(key)] = self.facedata[key]
-
-        for key in self.edgedata:
-            edgedata[repr(key)] = self.edgedata[key]
-
-        data = {'attributes': self.attributes,
-                'dva': self.default_vertex_attributes,
-                'dea': self.default_edge_attributes,
-                'dfa': self.default_face_attributes,
-                'vertex': self.vertex,
-                'face': self.face,
-                'facedata': self.facedata,
-                'edgedata': edgedata,
-                'max_int_key': self._max_int_key,
-                'max_int_fkey': self._max_int_fkey}
-
-        return data
+        data = {
+            'attributes': self.attributes,
+            'dva': self.default_vertex_attributes,
+            'dea': self.default_edge_attributes,
+            'dfa': self.default_face_attributes,
+            'vertex': self.vertex,
+            'face': self.face,
+            'facedata': self.facedata,
+        }
+        version = LooseVersion(compas.__version__)
+        if version < LooseVersion('0.16.5'):
+            data['edgedata'] = {repr(key): self.edgedata[key] for key in self.edgedata}
+            data['max_int_key'] = self._max_vertex
+            data['max_int_fkey'] = self._max_face
+            return data
+        data['edgedata'] = self.edgedata
+        data['max_vertex'] = self._max_vertex
+        data['max_face'] = self._max_face
+        return {
+            'compas': version.vstring.split('-')[0],
+            'datatype': self.dtype,
+            'data': data
+        }
 
     @data.setter
     def data(self, data):
-        attributes = data['attributes']
-        dva = data.get('dva') or {}
-        dfa = data.get('dfa') or {}
-        dea = data.get('dea') or {}
-        vertex = data.get('vertex') or {}
-        face = data.get('face') or {}
-        facedata = data.get('facedata') or {}
-        edgedata = data.get('edgedata') or {}
-        max_int_key = data.get('max_int_key', -1)
-        max_int_fkey = data.get('max_int_fkey', -1)
-
-        self.attributes.update(attributes)
-        self.default_vertex_attributes.update(dva)
-        self.default_face_attributes.update(dfa)
-        self.default_edge_attributes.update(dea)
-
-        self.vertex = {}
-        self.face = {}
-        self.halfedge = {}
-        self.facedata = {}
-        self.edgedata = {}
-
-        for key, attr in iter(vertex.items()):
-            self.add_vertex(int(key), attr_dict=attr)
-
-        for fkey, vertices in iter(face.items()):
-            attr = facedata.get(fkey) or {}
-            # vertices = [int(k) for k in vertices]
-            self.add_face(vertices, fkey=int(fkey), attr_dict=attr)
-
-        for uv, attr in iter(edgedata.items()):
-            self.edgedata[literal_eval(uv)] = attr or {}
-
-        self._max_int_key = max_int_key
-        self._max_int_fkey = max_int_fkey
-
-    # --------------------------------------------------------------------------
-    # from/to
-    # --------------------------------------------------------------------------
-
-    @classmethod
-    def from_data(cls, data):
-        """Construct a mesh from structured data.
-
-        Parameters
-        ----------
-        data : dict
-            The data dictionary.
-
-        Returns
-        -------
-        object
-            An object of the type of ``cls``.
-
-        Notes
-        -----
-        This constructor method is meant to be used in conjunction with the
-        corresponding *to_data* method.
-
-        """
-        mesh = cls()
-        mesh.data = data
-        return mesh
-
-    def to_data(self):
-        """Returns a dictionary of structured data representing the mesh.
-
-        Returns
-        -------
-        dict
-            The structured data.
-
-        Notes
-        ----
-        This method produces the data that can be used in conjunction with the
-        corresponding *from_data* class method.
-        """
-        return self.data
-
-    @classmethod
-    def from_json(cls, filepath):
-        """Construct a datastructure from structured data contained in a json file.
-
-        Parameters
-        ----------
-        filepath : str
-            The path to the json file.
-
-        Returns
-        -------
-        object
-            An object of the type of ``cls``.
-
-        Notes
-        -----
-        This constructor method is meant to be used in conjunction with the
-        corresponding *to_json* method.
-        """
-        with open(filepath, 'r') as fp:
-            data = json.load(fp)
-        mesh = cls()
-        mesh.data = data
-        return mesh
-
-    def to_json(self, filepath, pretty=False):
-        """Serialise the structured data representing the data structure to json.
-
-        Parameters
-        ----------
-        filepath : str
-            The path to the json file.
-        """
-        with open(filepath, 'w+') as f:
-            if pretty:
-                json.dump(self.data, f, sort_keys=True, indent=4)
-            else:
-                json.dump(self.data, f)
-
-    @classmethod
-    def from_pickle(cls, filepath):
-        """Construct a mesh from serialised data contained in a pickle file.
-
-        Parameters
-        ----------
-        filepath : str
-            The path to the pickle file.
-
-        Returns
-        -------
-        object
-            An object of type ``cls``.
-
-        Notes
-        -----
-        This constructor method is meant to be used in conjunction with the
-        corresponding *to_pickle* method.
-        """
-        with open(filepath, 'rb') as fo:
-            data = pickle.load(fo)
-        o = cls()
-        o.data = data
-        return o
-
-    def to_pickle(self, filepath):
-        """Serialise the structured data representing the mesh to a pickle file.
-
-        Parameters
-        ----------
-        filepath : str
-            The path to the pickle file.
-        """
-        with open(filepath, 'wb+') as f:
-            pickle.dump(self.data, f, protocol=pickle.HIGHEST_PROTOCOL)
+        if 'compas' in data:
+            version = LooseVersion(compas.__version__)
+            if version < LooseVersion('0.16.5'):
+                raise Exception('The data was generated with an incompatible newer version of COMPAS: {}'.format(version.vstring.split('-')[0]))
+            # dtype = data['dtype']
+            data = data['data']
+            attributes = data['attributes']
+            dva = data.get('dva') or {}
+            dfa = data.get('dfa') or {}
+            dea = data.get('dea') or {}
+            vertex = data.get('vertex') or {}
+            face = data.get('face') or {}
+            facedata = data.get('facedata') or {}
+            edgedata = data.get('edgedata') or {}
+            max_vertex = data.get('max_vertex', -1)
+            max_face = data.get('max_face', -1)
+            self.attributes.update(attributes)
+            self.default_vertex_attributes.update(dva)
+            self.default_face_attributes.update(dfa)
+            self.default_edge_attributes.update(dea)
+            self.vertex = {}
+            self.face = {}
+            self.halfedge = {}
+            self.facedata = {}
+            self.edgedata = {}
+            # this could be handled by the schema
+            # but will not work in IronPython
+            for key, attr in iter(vertex.items()):
+                self.add_vertex(int(key), attr_dict=attr)
+            for fkey, vertices in iter(face.items()):
+                attr = facedata.get(fkey) or {}
+                self.add_face(vertices, fkey=int(fkey), attr_dict=attr)
+            for uv, attr in iter(edgedata.items()):
+                self.edgedata[uv] = attr or {}
+            self._max_vertex = max_vertex
+            self._max_face = max_face
+        else:
+            attributes = data['attributes']
+            dva = data.get('dva') or {}
+            dfa = data.get('dfa') or {}
+            dea = data.get('dea') or {}
+            vertex = data.get('vertex') or {}
+            face = data.get('face') or {}
+            facedata = data.get('facedata') or {}
+            edgedata = data.get('edgedata') or {}
+            max_vertex = data.get('max_int_key', -1)
+            max_face = data.get('max_int_fkey', -1)
+            self.attributes.update(attributes)
+            self.default_vertex_attributes.update(dva)
+            self.default_face_attributes.update(dfa)
+            self.default_edge_attributes.update(dea)
+            self.vertex = {}
+            self.face = {}
+            self.halfedge = {}
+            self.facedata = {}
+            self.edgedata = {}
+            # this could be handled by the schema
+            # but will not work in IronPython
+            for key, attr in iter(vertex.items()):
+                self.add_vertex(int(key), attr_dict=attr)
+            for fkey, vertices in iter(face.items()):
+                attr = facedata.get(fkey) or {}
+                self.add_face(vertices, fkey=int(fkey), attr_dict=attr)
+            for edge, attr in iter(edgedata.items()):
+                key = "-".join(map(str, sorted(literal_eval(edge))))
+                if key not in self.edgedata:
+                    self.edgedata[key] = {}
+                if attr:
+                    self.edgedata[key].update(attr)
+            self._max_vertex = max_vertex
+            self._max_face = max_face
 
     # --------------------------------------------------------------------------
     # helpers
     # --------------------------------------------------------------------------
 
-    def copy(self, cls=None):
-        """Make an independent copy of the mesh object.
-
-        Parameters
-        ----------
-        cls : compas.datastructures.Mesh, optional
-            The type of mesh to return.
-            Defaults to the type of the current mesh.
-
-        Returns
-        -------
-        Mesh
-            A separate, but identical mesh object.
-        """
-        if not cls:
-            cls = type(self)
-        return cls.from_data(deepcopy(self.data))
-
     def clear(self):
         """Clear all the mesh data."""
         del self.vertex
@@ -388,8 +298,8 @@ class HalfEdge(Datastructure):
         self.halfedge = {}
         self.face = {}
         self.facedata = {}
-        self._max_int_key = -1
-        self._max_int_fkey = -1
+        self._max_vertex = -1
+        self._max_face = -1
 
     def get_any_vertex(self):
         """Get the identifier of a random vertex.
@@ -464,6 +374,8 @@ class HalfEdge(Datastructure):
         """
         return {key: index for index, key in enumerate(self.vertices())}
 
+    vertex_index = key_index
+
     def index_key(self):
         """Returns a dictionary that maps the indices of a vertex list to
         keys in a vertex dictionary.
@@ -476,6 +388,8 @@ class HalfEdge(Datastructure):
         """
         return dict(enumerate(self.vertices()))
 
+    index_vertex = index_key
+
     # --------------------------------------------------------------------------
     # builders
     # --------------------------------------------------------------------------
@@ -520,9 +434,9 @@ class HalfEdge(Datastructure):
         0
         """
         if key is None:
-            key = self._max_int_key = self._max_int_key + 1
-        if key > self._max_int_key:
-            self._max_int_key = key
+            key = self._max_vertex = self._max_vertex + 1
+        if key > self._max_vertex:
+            self._max_vertex = key
         key = int(key)
         if key not in self.vertex:
             self.vertex[key] = {}
@@ -576,9 +490,9 @@ class HalfEdge(Datastructure):
         if len(vertices) < 3:
             return
         if fkey is None:
-            fkey = self._max_int_fkey = self._max_int_fkey + 1
-        if fkey > self._max_int_fkey:
-            self._max_int_fkey = fkey
+            fkey = self._max_face = self._max_face + 1
+        if fkey > self._max_face:
+            self._max_face = fkey
         attr = attr_dict or {}
         attr.update(kwattr)
         self.face[fkey] = vertices
@@ -623,19 +537,25 @@ class HalfEdge(Datastructure):
                 del self.facedata[fkey]
         for nbr in nbrs:
             del self.halfedge[nbr][key]
-            if (nbr, key) in self.edgedata:
-                del self.edgedata[nbr, key]
-            if (key, nbr) in self.edgedata:
-                del self.edgedata[key, nbr]
+            edge = "-".join(map(str, sorted([nbr, key])))
+            if edge in self.edgedata:
+                del self.edgedata[edge]
+            # if (nbr, key) in self.edgedata:
+            #     del self.edgedata[nbr, key]
+            # if (key, nbr) in self.edgedata:
+            #     del self.edgedata[key, nbr]
         for nbr in nbrs:
             for n in self.vertex_neighbors(nbr):
                 if self.halfedge[nbr][n] is None and self.halfedge[n][nbr] is None:
                     del self.halfedge[nbr][n]
                     del self.halfedge[n][nbr]
-                    if (nbr, n) in self.edgedata:
-                        del self.edgedata[nbr, n]
-                    if (n, nbr) in self.edgedata:
-                        del self.edgedata[n, nbr]
+                    edge = "-".join(map(str, sorted([nbr, n])))
+                    if edge in self.edgedata:
+                        del self.edgedata[edge]
+                    # if (nbr, n) in self.edgedata:
+                    #     del self.edgedata[nbr, n]
+                    # if (n, nbr) in self.edgedata:
+                    #     del self.edgedata[n, nbr]
         del self.halfedge[key]
         del self.vertex[key]
 
@@ -662,15 +582,14 @@ class HalfEdge(Datastructure):
             if self.halfedge[v][u] is None:
                 del self.halfedge[u][v]
                 del self.halfedge[v][u]
-                if (u, v) in self.edgedata:
-                    del self.edgedata[u, v]
-                if (v, u) in self.edgedata:
-                    del self.edgedata[v, u]
+                edge = "-".join(map(str, sorted([u, v])))
+                if edge in self.edgedata:
+                    del self.edgedata[edge]
         del self.face[fkey]
         if fkey in self.facedata:
             del self.facedata[fkey]
 
-    def cull_vertices(self):
+    def remove_unused_vertices(self):
         """Remove all unused vertices from the mesh object.
         """
         for u in list(self.vertices()):
@@ -681,6 +600,8 @@ class HalfEdge(Datastructure):
                     del self.vertex[u]
                     del self.halfedge[u]
 
+    cull_vertices = remove_unused_vertices
+
     # --------------------------------------------------------------------------
     # accessors
     # --------------------------------------------------------------------------
@@ -1456,12 +1377,12 @@ class HalfEdge(Datastructure):
         attr_dict.update(kwattr)
         self.default_edge_attributes.update(attr_dict)
 
-    def edge_attribute(self, key, name, value=None):
+    def edge_attribute(self, edge, name, value=None):
         """Get or set an attribute of an edge.
 
         Parameters
         ----------
-        key : 2-tuple of int
+        edge : 2-tuple of int
             The identifier of the edge as a pair of vertex identifiers.
         name : str
             The name of the attribute.
@@ -1479,29 +1400,26 @@ class HalfEdge(Datastructure):
         KeyError
             If the edge does not exist.
         """
-        u, v = key
+        u, v = edge
         if u not in self.halfedge or v not in self.halfedge[u]:
-            raise KeyError(key)
+            raise KeyError(edge)
+        key = "-".join(map(str, sorted(edge)))
         if value is not None:
-            if (u, v) not in self.edgedata:
-                self.edgedata[u, v] = {}
-            if (v, u) not in self.edgedata:
-                self.edgedata[v, u] = {}
-            self.edgedata[u, v][name] = self.edgedata[v, u][name] = value
+            if key not in self.edgedata:
+                self.edgedata[key] = {}
+            self.edgedata[key][name] = value
             return
-        if (u, v) in self.edgedata and name in self.edgedata[u, v]:
-            return self.edgedata[u, v][name]
-        if (v, u) in self.edgedata and name in self.edgedata[v, u]:
-            return self.edgedata[v, u][name]
+        if key in self.edgedata and name in self.edgedata[key]:
+            return self.edgedata[key][name]
         if name in self.default_edge_attributes:
             return self.default_edge_attributes[name]
 
-    def unset_edge_attribute(self, key, name):
+    def unset_edge_attribute(self, edge, name):
         """Unset the attribute of an edge.
 
         Parameters
         ----------
-        key : tuple of int
+        edge : tuple of int
             The edge identifier.
         name : str
             The name of the attribute.
@@ -1516,23 +1434,19 @@ class HalfEdge(Datastructure):
         Unsetting the value of an edge attribute implicitly sets it back to the value
         stored in the default edge attribute dict.
         """
-        u, v = key
+        u, v = edge
         if u not in self.halfedge or v not in self.halfedge[u]:
-            raise KeyError(key)
-        if key in self.edgedata:
-            if name in self.edgedata[key]:
-                del self.edgedata[key][name]
-        key = v, u
-        if key in self.edgedata:
-            if name in self.edgedata[key]:
-                del self.edgedata[key][name]
+            raise KeyError(edge)
+        key = "-".join(map(str, sorted(edge)))
+        if key in self.edgedata and name in self.edgedata[key]:
+            del self.edgedata[key][name]
 
-    def edge_attributes(self, key, names=None, values=None):
+    def edge_attributes(self, edge, names=None, values=None):
         """Get or set multiple attributes of an edge.
 
         Parameters
         ----------
-        key : 2-tuple of int
+        edge : 2-tuple of int
             The identifier of the edge.
         names : list, optional
             A list of attribute names.
@@ -1553,22 +1467,23 @@ class HalfEdge(Datastructure):
         KeyError
             If the edge does not exist.
         """
-        u, v = key
+        u, v = edge
         if u not in self.halfedge or v not in self.halfedge[u]:
-            raise KeyError(key)
+            raise KeyError(edge)
         if values is not None:
             # use it as a setter
             for name, value in zip(names, values):
-                self.edge_attribute(key, name, value)
+                self.edge_attribute(edge, name, value)
             return
         # use it as a getter
         if not names:
+            key = "-".join(map(str, sorted(edge)))
             # get the entire attribute dict
             return EdgeAttributeView(self.default_edge_attributes, self.edgedata.setdefault(key, {}))
         # get only the values of the named attributes
         values = []
         for name in names:
-            value = self.edge_attribute(key, name)
+            value = self.edge_attribute(edge, name)
             values.append(value)
         return values
 
@@ -1582,7 +1497,7 @@ class HalfEdge(Datastructure):
         value : obj, optional
             The value of the attribute.
             Default is ``None``.
-        keys : list of 2-tuple of int, optional
+        keys : list of edges, optional
             A list of edge identifiers.
 
         Returns
@@ -1596,13 +1511,12 @@ class HalfEdge(Datastructure):
         KeyError
             If any of the edges does not exist.
         """
-        if not keys:
-            keys = self.edges()
+        edges = keys or self.edges()
         if value is not None:
-            for key in keys:
-                self.edge_attribute(key, name, value)
+            for edge in edges:
+                self.edge_attribute(edge, name, value)
             return
-        return [self.edge_attribute(key, name) for key in keys]
+        return [self.edge_attribute(edge, name) for edge in edges]
 
     def edges_attributes(self, names=None, values=None, keys=None):
         """Get or set multiple attributes of multiple edges.
@@ -1615,7 +1529,7 @@ class HalfEdge(Datastructure):
         values : list of obj, optional
             The values of the attributes.
             Default is ``None``.
-        keys : list of 2-tuple of int, optional
+        keys : list of edges, optional
             A list of edge identifiers.
 
         Returns
@@ -1632,18 +1546,27 @@ class HalfEdge(Datastructure):
         KeyError
             If any of the edges does not exist.
         """
-        if not keys:
-            keys = self.edges()
+        edges = keys or self.edges()
         if values is not None:
-            for key in keys:
-                self.edge_attributes(key, names, values)
+            for edge in edges:
+                self.edge_attributes(edge, names, values)
             return
-        return [self.edge_attributes(key, names) for key in keys]
+        return [self.edge_attributes(edge, names) for edge in edges]
 
     # --------------------------------------------------------------------------
     # mesh info
     # --------------------------------------------------------------------------
 
+    def summary(self):
+        """Print a summary of the mesh.
+
+        Returns
+        -------
+        str
+        """
+        tpl = "\n".join(["{} summary", "=" * (len(self.name) + len(" summary")), "- vertices: {}", "- edges: {}", "- faces: {}"])
+        return tpl.format(self.name, self.number_of_vertices(), self.number_of_edges(), self.number_of_faces())
+
     def number_of_vertices(self):
         """Count the number of vertices in the mesh."""
         return len(list(self.vertices()))
@@ -2202,9 +2125,131 @@ class HalfEdge(Datastructure):
     # polyedge topology
     # --------------------------------------------------------------------------
 
-    # face strips?
-    # edge chains?
-    # ...?
+    def edge_loop(self, edge):
+        """Find all edges on the same loop as a given edge.
+
+        Parameters
+        ----------
+        edge : tuple of int
+            The identifier of the starting edge.
+
+        Returns
+        -------
+        list of tuple of int
+            The edges on the same loop as the given edge.
+        """
+        if self.is_edge_on_boundary(*edge):
+            return self._edge_loop_on_boundary(edge)
+        edges = []
+        current, previous = edge
+        edges.append((previous, current))
+        while True:
+            if current == edge[1]:
+                break
+            nbrs = self.vertex_neighbors(current, ordered=True)
+            if len(nbrs) != 4:
+                break
+            i = nbrs.index(previous)
+            previous = current
+            current = nbrs[i - 2]
+            edges.append((previous, current))
+        edges[:] = [(u, v) for v, u in edges[::-1]]
+        if edges[0][0] == edges[-1][1]:
+            return edges
+        previous, current = edge
+        while True:
+            nbrs = self.vertex_neighbors(current, ordered=True)
+            if len(nbrs) != 4:
+                break
+            i = nbrs.index(previous)
+            previous = current
+            current = nbrs[i - 2]
+            edges.append((previous, current))
+        return edges
+
+    def _edge_loop_on_boundary(self, uv):
+        """Find all edges on the same loop as a given edge on the boundary."""
+        edges = []
+        current, previous = uv
+        edges.append((previous, current))
+        while True:
+            if current == uv[1]:
+                break
+            nbrs = self.vertex_neighbors(current)
+            if len(nbrs) == 2:
+                break
+            nbr = None
+            for temp in nbrs:
+                if temp == previous:
+                    continue
+                if self.is_edge_on_boundary(current, temp):
+                    nbr = temp
+                    break
+            if nbr is None:
+                break
+            previous, current = current, nbr
+            edges.append((previous, current))
+        edges[:] = [(u, v) for v, u in edges[::-1]]
+        if edges[0][0] == edges[-1][1]:
+            return edges
+        previous, current = uv
+        while True:
+            nbrs = self.vertex_neighbors(current)
+            if len(nbrs) == 2:
+                break
+            nbr = None
+            for temp in nbrs:
+                if temp == previous:
+                    continue
+                if self.is_edge_on_boundary(current, temp):
+                    nbr = temp
+                    break
+            if nbr is None:
+                break
+            previous, current = current, nbr
+            edges.append((previous, current))
+        return edges
+
+    def edge_strip(self, edge):
+        """Find all edges on the same strip as a given edge.
+
+        Parameters
+        ----------
+        edge : tuple of int
+            The identifier of the starting edge.
+
+        Returns
+        -------
+        list of tuple of int
+            The edges on the same strip as the given edge.
+        """
+        edges = []
+        v, u = edge
+        while True:
+            edges.append((u, v))
+            face = self.halfedge[u][v]
+            if face is None:
+                break
+            vertices = self.face_vertices(face)
+            if len(vertices) != 4:
+                break
+            i = vertices.index(u)
+            u = vertices[i - 1]
+            v = vertices[i - 2]
+        edges[:] = [(u, v) for v, u in edges[::-1]]
+        u, v = edge
+        while True:
+            face = self.halfedge[u][v]
+            if face is None:
+                break
+            vertices = self.face_vertices(face)
+            if len(vertices) != 4:
+                break
+            i = vertices.index(u)
+            u = vertices[i - 1]
+            v = vertices[i - 2]
+            edges.append((u, v))
+        return edges
 
     # --------------------------------------------------------------------------
     # face topology
@@ -2484,22 +2529,6 @@ class HalfEdge(Datastructure):
         else:
             return False
 
-    # --------------------------------------------------------------------------
-    # mesh geometry
-    # --------------------------------------------------------------------------
-
-    # --------------------------------------------------------------------------
-    # vertex geometry
-    # --------------------------------------------------------------------------
-
-    # --------------------------------------------------------------------------
-    # edge geometry
-    # --------------------------------------------------------------------------
-
-    # --------------------------------------------------------------------------
-    # face geometry
-    # --------------------------------------------------------------------------
-
     # --------------------------------------------------------------------------
     # boundary
     # --------------------------------------------------------------------------
