@@ -41,7 +41,10 @@ def clear():
             bpy.data.collections.remove(block)
 
 
-__version__ = '0.16.5'
+__version__ = '0.17.1'
 
 
 __all__ = [name for name in dir() if not name.startswith('_')]
+__all_plugins__ = [
+    # 'compas_blender.geometry.booleans',
+]
