@@ -3,7 +3,7 @@ from __future__ import absolute_import
 from __future__ import division
 
 # import compas_rhino
-from compas_rhino.artists._primitiveartist import PrimitiveArtist
+from ._primitiveartist import PrimitiveArtist
 
 
 __all__ = ['PlaneArtist']
